package com.ippon.error.domain;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Parent exception used in your application. Those exceptions will be resolved as human readable errors.
 *
 * <p>
 * You can use this implementation directly:
 * </p>
 *
 * <p>
 * <code>
 *     <pre>
 *       MyAppException.builder(StandardMessages.USER_MANDATORY)
 *          .argument("key", "value")
 *          .argument("other", "test")
 *          .status(ErrorsHttpStatus.BAD_REQUEST)
 *          .message("Error message")
 *          .cause(new RuntimeException())
 *          .build();
 *     </pre>
 *   </code>
 * </p>
 *
 * <p>
 * Or make extension exceptions:
 * </p>
 *
 * <p>
 * <code>
 *     <pre>
 *       public class MissingMandatoryValueException extends MyAppException {
 *
 *         public MissingMandatoryValueException(MyAppMessage myAppMessage, String fieldName) {
 *           this(builder(myAppMessage, fieldName, defaultMessage(fieldName)));
 *         }
 *
 *         protected MissingMandatoryValueException(MyAppExceptionBuilder builder) {
 *           super(builder);
 *         }
 *
 *         private static MyAppExceptionBuilder builder(MyAppMessage myAppMessage, String fieldName, String message) {
 *           return MyAppException.builder(myAppMessage)
 *               .status(ErrorsHttpStatus.INTERNAL_SERVER_ERROR)
 *               .argument("field", fieldName)
 *               .message(message);
 *         }
 *
 *         private static String defaultMessage(String fieldName) {
 *           return "The field \"" + fieldName + "\" is mandatory and wasn't set";
 *         }
 *       }
 *     </pre>
 *   </code>
 * </p>
 */
public class MyAppException extends RuntimeException {
  private final Map<String, String> arguments;
  private final ErrorStatus status;
  private final MyAppMessage myAppMessage;

  protected MyAppException(MyAppExceptionBuilder builder) {
    super(getMessage(builder), getCause(builder));
    arguments = getArguments(builder);
    status = getStatus(builder);
    myAppMessage = getMyAppMessage(builder);
  }

  private static String getMessage(MyAppExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.message;
  }

  private static Throwable getCause(MyAppExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.cause;
  }

  private static Map<String, String> getArguments(MyAppExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return Collections.unmodifiableMap(builder.arguments);
  }

  private static ErrorStatus getStatus(MyAppExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.status;
  }

  private static MyAppMessage getMyAppMessage(MyAppExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.myAppMessage;
  }

  public static MyAppExceptionBuilder builder(MyAppMessage message) {
    return new MyAppExceptionBuilder(message);
  }

  public Map<String, String> getArguments() {
    return arguments;
  }

  public ErrorStatus getStatus() {
    return status;
  }

  public MyAppMessage getMyAppMessage() {
    return myAppMessage;
  }

  public static class MyAppExceptionBuilder {
    private final Map<String, String> arguments = new HashMap<>();
    private String message;
    private ErrorStatus status;
    private MyAppMessage myAppMessage;
    private Throwable cause;

    public MyAppExceptionBuilder(MyAppMessage myAppMessage) {
      this.myAppMessage = myAppMessage;
    }

    public MyAppExceptionBuilder argument(String key, Object value) {
      arguments.put(key, getStringValue(value));

      return this;
    }

    private String getStringValue(Object value) {
      if (value == null) {
        return "null";
      }

      return value.toString();
    }

    public MyAppExceptionBuilder status(ErrorStatus status) {
      this.status = status;

      return this;
    }

    public MyAppExceptionBuilder message(String message) {
      this.message = message;

      return this;
    }

    public MyAppExceptionBuilder cause(Throwable cause) {
      this.cause = cause;

      return this;
    }

    public MyAppException build() {
      return new MyAppException(this);
    }
  }
}
