package com.ippon.pagination.domain;

import com.ippon.error.domain.Assert;
import java.util.Collections;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;

public class MyAppPage<T> {
  private static final int MINIMAL_PAGE_COUNT = 1;

  private final List<T> content;
  private final int currentPage;
  private final int pageSize;
  private final long totalElementsCount;
  private final int pageCount;

  private MyAppPage(MyAppPageBuilder<T> builder) {
    content = buildContent(builder.content);
    currentPage = builder.currentPage;
    pageSize = buildPageSize(builder.pageSize);
    totalElementsCount = buildTotalElementsCount(builder.totalElementsCount);
    pageCount = buildPageCount();
  }

  private List<T> buildContent(List<T> content) {
    if (content == null) {
      return List.of();
    }

    return Collections.unmodifiableList(content);
  }

  private int buildPageSize(int pageSize) {
    if (pageSize == -1) {
      return content.size();
    }

    return pageSize;
  }

  private long buildTotalElementsCount(long totalElementsCount) {
    if (totalElementsCount == -1) {
      return content.size();
    }

    return totalElementsCount;
  }

  public int buildPageCount() {
    if (totalElementsCount > 0) {
      return (int) Math.ceil(totalElementsCount / (float) pageSize);
    }

    return MINIMAL_PAGE_COUNT;
  }

  public static <T> MyAppPageBuilder<T> builder(List<T> content) {
    return new MyAppPageBuilder<>(content);
  }

  public static <T> MyAppPage<T> singlePage(List<T> content) {
    return builder(content).build();
  }

  public List<T> getContent() {
    return content;
  }

  public int getCurrentPage() {
    return currentPage;
  }

  public int getPageSize() {
    return pageSize;
  }

  public long getTotalElementsCount() {
    return totalElementsCount;
  }

  public int getPageCount() {
    return pageCount;
  }

  public boolean isNotLast() {
    return currentPage + 1 != getPageCount();
  }

  public <R> MyAppPage<R> map(Function<T, R> mapper) {
    Assert.notNull("mapper", mapper);

    return builder(getContent().stream().map(mapper).collect(Collectors.toList()))
      .currentPage(currentPage)
      .pageSize(pageSize)
      .totalElementsCount(totalElementsCount)
      .build();
  }

  public static class MyAppPageBuilder<T> {
    private final List<T> content;
    private int currentPage;
    private int pageSize = -1;
    private long totalElementsCount = -1;

    private MyAppPageBuilder(List<T> content) {
      this.content = content;
    }

    public MyAppPageBuilder<T> pageSize(int pageSize) {
      this.pageSize = pageSize;

      return this;
    }

    public MyAppPageBuilder<T> currentPage(int currentPage) {
      this.currentPage = currentPage;

      return this;
    }

    public MyAppPageBuilder<T> totalElementsCount(long totalElementsCount) {
      this.totalElementsCount = totalElementsCount;

      return this;
    }

    public MyAppPage<T> build() {
      return new MyAppPage<>(this);
    }
  }
}
